#include "HeartBeatHelper.h"

#include <iostream>
#include <cstdio>
#include <cstring>

// rc4
static int m_sBox[256];
static int rc4_val_a, rc4_val_b;
static unsigned char swap;

// MD5
static unsigned char PADDING[64] = {
        0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

void hb_md5_init(HB_MD5CTX *md5_context) {
    md5_context->i[0] = md5_context->i[1] = (unsigned int) 0;

    md5_context->buf[0] = 0x67452301L;
    md5_context->buf[1] = 0xefcdab89L;
    md5_context->buf[2] = 0x98badcfeL;
    md5_context->buf[3] = 0x10325476L;
}

void hb_md5_transform(unsigned int *buf, const unsigned int *in) {
    register unsigned int a = buf[0], b = buf[1], c = buf[2], d = buf[3];

    /* Round 1 */
#define S11 7
#define S12 12
#define S13 17
#define S14 22
    FF(a, b, c, d, in[0], S11, 0xD76AA478L); /* 1 */
    FF(d, a, b, c, in[1], S12, 0xE8C7B756L); /* 2 */
    FF(c, d, a, b, in[2], S13, 0x242070DBL); /* 3 */
    FF(b, c, d, a, in[3], S14, 0xC1BDCEEEL); /* 4 */
    FF(a, b, c, d, in[4], S11, 0xF57C0FAFL); /* 5 */
    FF(d, a, b, c, in[5], S12, 0x4787C62AL); /* 6 */
    FF(c, d, a, b, in[6], S13, 0xA8304613L); /* 7 */
    FF(b, c, d, a, in[7], S14, 0xFD469501L); /* 8 */
    FF(a, b, c, d, in[8], S11, 0x698098D8L); /* 9 */
    FF(d, a, b, c, in[9], S12, 0x8B44F7AFL); /* 10 */
    FF(c, d, a, b, in[10], S13, 0xFFFF5BB1L); /* 11 */
    FF(b, c, d, a, in[11], S14, 0x895CD7BEL); /* 12 */
    FF(a, b, c, d, in[12], S11, 0x6B901122L); /* 13 */
    FF(d, a, b, c, in[13], S12, 0xFD987193L); /* 14 */
    FF(c, d, a, b, in[14], S13, 0xA679438EL); /* 15 */
    FF(b, c, d, a, in[15], S14, 0x49B40821L); /* 16 */

    /* Round 2 */
#define S21 5
#define S22 9
#define S23 14
#define S24 20
    GG(a, b, c, d, in[1], S21, 0xF61E2562L); /* 17 */
    GG(d, a, b, c, in[6], S22, 0xC040B340L); /* 18 */
    GG(c, d, a, b, in[11], S23, 0x265E5A51L); /* 19 */
    GG(b, c, d, a, in[0], S24, 0xE9B6C7AAL); /* 20 */
    GG(a, b, c, d, in[5], S21, 0xD62F105DL); /* 21 */
    GG(d, a, b, c, in[10], S22, 0x02441453L); /* 22 */
    GG(c, d, a, b, in[15], S23, 0xD8A1E681L); /* 23 */
    GG(b, c, d, a, in[4], S24, 0xE7D3FBC8L); /* 24 */
    GG(a, b, c, d, in[9], S21, 0x21E1CDE6L); /* 25 */
    GG(d, a, b, c, in[14], S22, 0xC33707D6L); /* 26 */
    GG(c, d, a, b, in[3], S23, 0xF4D50D87L); /* 27 */
    GG(b, c, d, a, in[8], S24, 0x455A14EDL); /* 28 */
    GG(a, b, c, d, in[13], S21, 0xA9E3E905L); /* 29 */
    GG(d, a, b, c, in[2], S22, 0xFCEFA3F8L); /* 30 */
    GG(c, d, a, b, in[7], S23, 0x676F02D9L); /* 31 */
    GG(b, c, d, a, in[12], S24, 0x8D2A4C8AL); /* 32 */

    /* Round 3 */
#define S31 4
#define S32 11
#define S33 16
#define S34 23
    HH(a, b, c, d, in[5], S31, 0xFFFA3942L); /* 33 */
    HH(d, a, b, c, in[8], S32, 0x8771F681L); /* 34 */
    HH(c, d, a, b, in[11], S33, 0x6D9D6122L); /* 35 */
    HH(b, c, d, a, in[14], S34, 0xFDE5380CL); /* 36 */
    HH(a, b, c, d, in[1], S31, 0xA4BEEA44L); /* 37 */
    HH(d, a, b, c, in[4], S32, 0x4BDECFA9L); /* 38 */
    HH(c, d, a, b, in[7], S33, 0xF6BB4B60L); /* 39 */
    HH(b, c, d, a, in[10], S34, 0xBEBFBC70L); /* 40 */
    HH(a, b, c, d, in[13], S31, 0x289B7EC6L); /* 41 */
    HH(d, a, b, c, in[0], S32, 0xEAA127FAL); /* 42 */
    HH(c, d, a, b, in[3], S33, 0xD4EF3085L); /* 43 */
    HH(b, c, d, a, in[6], S34, 0x04881D05L); /* 44 */
    HH(a, b, c, d, in[9], S31, 0xD9D4D039L); /* 45 */
    HH(d, a, b, c, in[12], S32, 0xE6DB99E5L); /* 46 */
    HH(c, d, a, b, in[15], S33, 0x1FA27CF8L); /* 47 */
    HH(b, c, d, a, in[2], S34, 0xC4AC5665L); /* 48 */

    /* Round 4 */
#define S41 6
#define S42 10
#define S43 15
#define S44 21
    II(a, b, c, d, in[0], S41, 0xF4292244L); /* 49 */
    II(d, a, b, c, in[7], S42, 0x432AFF97L); /* 50 */
    II(c, d, a, b, in[14], S43, 0xAB9423A7L); /* 51 */
    II(b, c, d, a, in[5], S44, 0xFC93A039L); /* 52 */
    II(a, b, c, d, in[12], S41, 0x655B59C3L); /* 53 */
    II(d, a, b, c, in[3], S42, 0x8F0CCC92L); /* 54 */
    II(c, d, a, b, in[10], S43, 0xFFEFF47DL); /* 55 */
    II(b, c, d, a, in[1], S44, 0x85845DD1L); /* 56 */
    II(a, b, c, d, in[8], S41, 0x6FA87E4FL); /* 57 */
    II(d, a, b, c, in[15], S42, 0xFE2CE6E0L); /* 58 */
    II(c, d, a, b, in[6], S43, 0xA3014314L); /* 59 */
    II(b, c, d, a, in[13], S44, 0x4E0811A1L); /* 60 */
    II(a, b, c, d, in[4], S41, 0xF7537E82L); /* 61 */
    II(d, a, b, c, in[11], S42, 0xBD3AF235L); /* 62 */
    II(c, d, a, b, in[2], S43, 0x2AD7D2BBL); /* 63 */
    II(b, c, d, a, in[9], S44, 0xEB86D391L); /* 64 */

    buf[0] += a;
    buf[1] += b;
    buf[2] += c;
    buf[3] += d;
}

void hb_md5_update(HB_MD5CTX *md5_context, const unsigned char *in, unsigned int len) {
    int mdi = (int) ((md5_context->i[0] >> 3) & 0x3F);
    unsigned int buf[16] = {0};

    if ((md5_context->i[0] + ((unsigned int) len << 3)) < md5_context->i[0])
        md5_context->i[1]++;
    md5_context->i[0] += ((unsigned int) len << 3);
    md5_context->i[1] += ((unsigned int) len >> 29);

    while (len--) {
        /* add new character to buffer, increment mdi */
        md5_context->in[mdi++] = *in++;

        /* transform if necessary */
        if (mdi == 0x40) {
            for (int i = 0, ii = 0; i < 16; ++i, ii += 4)
                buf[i] = (((unsigned int) md5_context->in[ii + 3]) << 24) |
                         (((unsigned int) md5_context->in[ii + 2]) << 16) |
                         (((unsigned int) md5_context->in[ii + 1]) << 8) |
                         ((unsigned int) md5_context->in[ii]);

            hb_md5_transform(md5_context->buf, buf);
            mdi = 0;
        }
    }
}

void hb_md5_final(unsigned char *digest, HB_MD5CTX *md5_context) {
    unsigned char pad_buf[64] = {
            0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
    };
    int mdi = (int) ((md5_context->i[0] >> 3) & 0x3F);
    unsigned int pad_len = (mdi < 56) ? (56 - mdi) : (120 - mdi);
    unsigned int in[16] = {0};
    in[14] = md5_context->i[0];
    in[15] = md5_context->i[1];

    hb_md5_update(md5_context, pad_buf, pad_len);

    /* append length in bits and transform */
    for (int i = 0, ii = 0; i < 14; i++, ii += 4)
        in[i] = (((unsigned int) md5_context->in[ii + 3]) << 24) |
                (((unsigned int) md5_context->in[ii + 2]) << 16) |
                (((unsigned int) md5_context->in[ii + 1]) << 8) |
                ((unsigned int) md5_context->in[ii]);
    hb_md5_transform(md5_context->buf, in);

    /* store buffer in digest */
    for (int i = 0, ii = 0; i < 4; i++, ii += 4) {
        unsigned int tmp = md5_context->buf[i];
        digest[ii] = (unsigned char) (tmp & 0xFF);
        digest[ii + 1] = (unsigned char) ((tmp >> 8) & 0xFF);
        digest[ii + 2] = (unsigned char) ((tmp >> 16) & 0xFF);
        digest[ii + 3] = (unsigned char) ((tmp >> 24) & 0xFF);
    }
}

// RC4
void rc4_init(unsigned char *pKey, unsigned int lenKey) {
    rc4_val_b = 0;
    for (rc4_val_a = 0; rc4_val_a < 256; rc4_val_a++) {
        m_sBox[rc4_val_a] = rc4_val_a;
    }
    for (rc4_val_a = 0; rc4_val_a < 256; rc4_val_a++) {
        rc4_val_b = (rc4_val_b + m_sBox[rc4_val_a] + pKey[rc4_val_a % lenKey]) % 256;
        swap = m_sBox[rc4_val_a];
        m_sBox[rc4_val_a] = m_sBox[rc4_val_b];
        m_sBox[rc4_val_b] = swap;
    }
}

void rc4_cipher(unsigned char *pData, unsigned long lenData) {
    int sBox[256];
    int i = 0, j = 0;
    long Offset;
    memcpy(sBox, m_sBox, 256 * sizeof(int));
    for (Offset = 0; Offset < lenData; Offset++) {
        i = (i + 1) % 256;
        j = (j + sBox[i]) % 256;
        swap = sBox[i];
        sBox[i] = sBox[j];
        sBox[j] = swap;
        pData[Offset] ^= sBox[(sBox[i] + sBox[j]) % 256];
    }
}


unsigned int *crc32_table() {
    static unsigned int s_crc32_tab[256] = {0};

    if (s_crc32_tab[1] == 0) {
        for (int i = 0; i < 256; i++) {
            unsigned int crc = i;
            for (int j = 0; j < 8; j++) {
                if (crc & 1)
                    crc = (crc >> 1) ^ 0xEDB88320;
                else
                    crc >>= 1;
            }
            s_crc32_tab[i] = crc;
        }
    }
    return s_crc32_tab;
}

unsigned int crc32(const void *in, int inlen) {

    unsigned int crc = 0;
    unsigned int *crc32_tab = crc32_table();
    const unsigned char *buffer = reinterpret_cast<const unsigned char *>(in);
    crc = 0xffffffff;
    for (int i = 0; i < inlen; i++) {
        crc = crc32_tab[(crc ^ buffer[i]) & 0xFF] ^ ((crc >> 8) & 0x00FFFFFF);
    }
    crc ^= 0xFFFFFFFF;
    return crc;
}

inline char get_index(char x) {
    //INITDATA
    const char *base64_lookup_array = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
    for (char i = 0; i < strlen(base64_lookup_array); i++) {
        if (x == base64_lookup_array[i])
            return i;
    }
    return 0;
}


void base64_decode(const void *in_buffer, size_t inlen, void *out_buffer, size_t *out_len) {

    unsigned int stitched_bytes = 0;
    int pad_count = 0;
    size_t calc_out_len = 0;
    const unsigned char *in = reinterpret_cast<const unsigned char *>(in_buffer);
    unsigned char *out = reinterpret_cast<unsigned char *>(out_buffer);

    for (size_t i = 0; i < inlen; i++) {
        stitched_bytes = 0;

        stitched_bytes = get_index(in[i]);
        stitched_bytes = stitched_bytes << 6;

        stitched_bytes |= get_index(in[++i]);
        stitched_bytes = stitched_bytes << 6;

        if (in[i + 1] == '=') pad_count = 2;

        stitched_bytes |= get_index(in[++i]);
        stitched_bytes = stitched_bytes << 6;

        if (in[i + 1] == '=' && pad_count != 2) {
            pad_count = 1;
        }

        stitched_bytes |= get_index(in[++i]);

        out[calc_out_len++] = (stitched_bytes >> 16) & 0xFF;

        if (pad_count == 2) {
            *out_len = calc_out_len;
            return;
        } else {
            out[calc_out_len++] = (stitched_bytes >> 8) & 0xFF;
            if (pad_count) {
                *out_len = calc_out_len;
                return;
            } else {
                out[calc_out_len++] = (stitched_bytes) & 0xFF;
            }
        }
    }
    *out_len = calc_out_len;
    return;
}


/**
 * 解密心跳数据
 * @param key    解密心跳数据的key
 * @param input  要解密的心跳数据
 * @return
 */
std::string DecodeHeartBeatData(std::string key, std::string input) {
    std::string ret = "";
    if (key.empty()) {
        std::cout << "[Error]: decodeHeartBeatData params key null!" << std::endl;
        return ret;
    }
    if (input.empty()) {
        std::cout << "[Error]: decodeHeartBeatData params input null!" << std::endl;
        return ret;
    }
    std::string::size_type pos = input.find("||");
    if (pos == input.npos) {
        return ret;
    }
    std::string base64_input = input.substr(0, pos);
    std::string ori_md5 = input.substr(pos + 2);
    char input_b64_dec[BASE64_DECODE_OUT_SIZE(base64_input.size())];
    size_t out_size = 0;
    base64_decode(base64_input.c_str(), base64_input.size(), input_b64_dec, &out_size);

    std::string ori_key;
    ori_key.append(key);
    ori_key.append("thiscontextenv");
    unsigned int crc32_val = crc32(ori_key.c_str(), ori_key.size());
    char rc4_key[16] = {0};
    snprintf(rc4_key, 9, "%x", crc32_val);

    rc4_init((unsigned char *) rc4_key, 8);
    std::string rc4_dec((char *) input_b64_dec, out_size);
    rc4_cipher((unsigned char *) rc4_dec.c_str(), out_size);

    if (rc4_dec.empty()) {
        std::cout << "[Failure] rc4 decode data" << std::endl;
        return ret;
    }
    std::string cal_md5;
    cal_md5.append(rc4_dec);
    cal_md5.append("thisisneteasesalt");

    char md5_str_bff[64] = {0};
    unsigned char md5_out_buffer[32] = {0};

    HB_MD5CTX ctx;
    hb_md5_init(&ctx);
    hb_md5_update(&ctx, (unsigned char *) cal_md5.c_str(), cal_md5.size());
    hb_md5_final(md5_out_buffer, &ctx);
    for (int i = 0; i < 16; ++i) {
        sprintf(md5_str_bff + i * 2, "%02x", md5_out_buffer[i]);
    }
    std::string dec_data_md5(md5_str_bff);
    if (ori_md5.compare(dec_data_md5) == 0) {
        std::cout << "[Success] decode heart beat data" << std::endl;
        return rc4_dec;
    }
    std::cout << "[Failure] decode heart beat data" << std::endl;
    return ret;
}